fitAndEvalSpline <- function( flightID, trj.df, knots, dist.vec, unit ) {
  
 if( unit == "FT" ) {
    dist.vec <- dist.vec / 0.3048
  }
  
  t.range <- range(trj.df$T)
  if ( t.range[2] - t.range[1] > 2 * trim.deltaT ) {
    min.trim <- t.range[1] + trim.deltaT
    max.trim <- t.range[2] - trim.deltaT
    tr.flag <- TRUE
  } else {
    tr.flag <- FALSE
  }

  df <- trj.df
  i <- 1
  
  while( i <= length(dist.vec) & nrow(df) > 4 ) { ## need at least 4 rows in df to fit a spline
    if(knots > nrow(df)) {
      knots <- nrow(df)
    }
    spl.X <- smooth.spline( df$T, df$X, w=df$WT, nknots=knots )
    spl.Y <- smooth.spline( df$T, df$Y, w=df$WT, nknots=knots )
    spl.Z <- smooth.spline( df$T, df$Z, w=df$WT, nknots=knots )
    fit.X <- predict( spl.X, df$T )
    fit.Y <- predict( spl.Y, df$T )
    fit.Z <- predict( spl.Z, df$T )
    
    dist <- dist.vec[i]
    w = abs( df$X - fit.X$y ) < dist & abs( df$Y - fit.Y$y ) < dist & abs( df$Z - fit.Z$y ) < dist 
    df <- df[w,]
    i <- i + 1
  }
  
  fit.3D <- data.frame( cbind( t(do.call( rbind, fit.X)), fit.Y$y, fit.Z$y) )
  colnames(fit.3D) <- c("T", "FITX", "FITY", "FITZ")

  if( tr.flag ) {
    w <- fit.3D[,"T"] >= min.trim & fit.3D[,"T"] <= max.trim
    tr.fit.3D <- data.frame(fit.3D[w,])
    tr.fit.df <- merge(trj.df, tr.fit.3D, by="T")
    if( nrow(tr.fit.df) > 0 ) {
      fit.TRRMSE <- sqrt( sum( (tr.fit.df$X - tr.fit.df$FITX)^2 + (tr.fit.df$Y - tr.fit.df$FITY)^2 + (tr.fit.df$Z - tr.fit.df$FITZ)^2 ) / nrow(tr.fit.df) )
    } else {
      fit.TRRMSE <- NA
    } 
  } else {
    fit.TRRMSE <- NA
  }
  
  trj.df   <- merge(trj.df, fit.3D, by="T")
  fit.RMSE <- sqrt( sum( (trj.df$X - trj.df$FITX)^2 + (trj.df$Y - trj.df$FITY)^2 + (trj.df$Z - trj.df$FITZ)^2 ) / nrow(trj.df) )
  
  return.list <- list( trj.df, fit.TRRMSE, fit.RMSE, spl.X, spl.Y, spl.Z )
  return( return.list )
  
}

## ------------------------------

pulse2trj <- function(pulse.dt, deltaT) {
  ## define time blocks of pulses
  minT <- min(pulse.dt[, "T"])
  minT <- floor(minT / deltaT) * deltaT ## aligns minT to increments of deltaT
  pulse.dt[, TBLOCK:= floor( (T - minT) / deltaT )] ## generates time blocks of returns labeled 0 to k.
  maxT <- ceiling(max(pulse.dt[, "T"]) / deltaT) * deltaT
  
  ## check if at least two eligible pulses are present in each timeblock. If only one, remove it
  tblBlock.df <- count(pulse.dt$TBLOCK)
  w <- tblBlock.df$freq > 1 ## you need at least two pulses to compute a platform location estimate
  pulse.dt <- subset(pulse.dt, TBLOCK %in% tblBlock.df$x[w]) 
  ## compute pulse weights
  pulse.dt[, WT:= exp(SEGLENFactor)*SEGLEN + exp(AngleFactor)*abs(SCANANGLE)]
  rm(w); gc()
  
  ## Convert pulses with negative SCANANGLE to negative weight
  pulse.dt[SCANANGLE < 0, WT:= -1.0 * WT]
  
  ## filter to two 'best' pulses per time Block
  ## .C returns a list of a number, a pulse-weight vector, a time block vector, and a flag vector.
  ## picking the fourth.
  pulse.dt[, PULSEFLAG:= .C( "filterTimeBlockPulses",
                             as.integer(nrow(pulse.dt)), as.double(pulse.dt$WT), as.integer(pulse.dt$TBLOCK),
                             as.integer(rep(-1, nrow(pulse.dt))) )[[4]] ]
  pulse.dt <- subset(pulse.dt, PULSEFLAG == 1, select=1:11)
  
  ## calculate cpa
  v <- rep( 0, nrow(pulse.dt)/2 )
  C.output <- .C( "cmpCPA",
                  as.integer( nrow(pulse.dt)/2 ),
                  as.double(pulse.dt$XLAST),  as.double(pulse.dt$YLAST),  as.double(pulse.dt$ZLAST),
                  as.double(pulse.dt$XFIRST), as.double(pulse.dt$YFIRST), as.double(pulse.dt$ZFIRST),
                  as.double(pulse.dt$T),      as.double(pulse.dt$WT),
                  as.double(v),               as.double(v),               as.double(v),
                  as.double(v),               as.double(v),               as.double(v),
                  as.double(SMALL.FLT) )
  trj.df <- as.data.frame(do.call(cbind, C.output[c(10:15)]))
  names(trj.df) <- c("X", "Y", "Z", "DIST", "T", "WT")
  
  return(trj.df)
}

## ------------------------------

exportSpline <- function(spl.X, spl.Y, spl.Z, minT, maxT, exportFrequency, exportType, flightID) {
  T.vec     <- seq(minT, maxT, by=1/exportFrequency)
  spline.dt <- data.table( X = round(predict(spl.X, T.vec)$y,2),
                           Y = round(predict(spl.Y, T.vec)$y,2),
                           Z = round(predict(spl.Z, T.vec)$y,2),
                           T = round(T.vec,4) )
  if( toupper(exportType) == "TEXT" ) {
    exportName <- paste0(getwd(), "/swath_", flightID, "_trj.csv")
    cat(exportName, "\n")
    print(getwd())
    write.csv(spline.dt, exportName, row.names=FALSE)
  }
  if( toupper(exportType) == "DBF" ) {
    exportName <- paste0("swath_", flightID, "_trj.dbf")
    write.dbf(spline.dt, exportName)
  }
}
